#include "config.h"
#ifdef HAVE_LIBCLIENT		/* OCI client library */

#ifdef HAVE_LIBFCGI
/* DO NOT #include <stdio.h> */
#include <fcgi_stdio.h>
#else
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>

/* Include OCI-specific headers. */
#include <oratypes.h>
#include <ocidfn.h>
#include <ociapr.h>
/* Many important constants are defined only in the "demo" include file. */
#include <ocidem.h>

static const char rcsid[] = "$Id: oci.c,v 1.11 1997/07/24 20:27:27 brlewis Exp $";

static char brl_nomem[] = "(error \"Out of memory.\")";


#ifndef NO_BRL_DEBUG
extern int debugfd;
#endif

/* poor man's profiling */
struct timeval starttime, endtime;
unsigned long usec_oopen=0, usec_oparse=0, usec_odescr=0, usec_odefin=0,
  usec_oexec=0, usec_ofetch=0, usec_brl_oci_writestring=0;

void
tstart()
{
  gettimeofday(&starttime);
  return;
}

unsigned long
tend()
{     
  gettimeofday(&endtime);
  return (unsigned long) (1000000 * (endtime.tv_sec - starttime.tv_sec)
			  + (endtime.tv_usec - starttime.tv_usec));
}

#define MAX_COLUMN_DESCRIPTION 256

struct brl_oci_connection
{
  char *b_conn_string;
  Lda_Def *b_lda;
  time_t b_lastused;
  char *b_hda;
};

static struct brl_oci_connection *conn_cache=NULL;
static int conn_cache_size=0;

struct brl_oci_outvar
{
  sword b_pos;			/* row position starting at 1 */

  /* for odescr() */
  sb4 b_dbsize;			/* max size of column (OCI ref 4-51) */
  sb2 b_dbtype;			/* internal datatype (OCI ref table 3-1) */
  sb1 b_cbuf[MAX_COLUMN_DESCRIPTION]; /* storage for column description */
  sb4 b_cbufl;			/* length of cbuf */
  sb4 b_dsize;			/* max size if returned as char string */
  sb2 b_prec;			/* precision of numeric items (OCI p. 3-3) */
  sb2 b_scale;			/* scale of numeric items */
  sb2 b_nullok;			/* non-zero iff nulls are permitted */

  /* for odefin() */
  ub1 *b_buf;			/* storage for oexfet() */
  sword b_bufl;			/* length of buf */
  sword b_ftype;		/* external datatype (OCI ref ch 3) */
  sb2 b_ind;			/* indicator for truncation or NULL value */
  ub2 b_rlen;			/* length of datum fetched */
  ub2 b_rcode;			/* return code for fetch */
};

/* maximum number of fields handled from SELECT output */
#define MAXFIELDS 1024
/* maximum size of column header handled from odescr() */
#define MAXFIELDHEADER 256

int
brl_oci_connect(conn_string)
     char *conn_string;
{
  int i, firstfree=-1;
  char sqlbuf[256], new_password[16];

  /* Reuse existing connection for this string */
  for(i=0; i < conn_cache_size; i++)
    {
      if (!conn_cache[i].b_conn_string) firstfree=i;
      else if (!strcmp(conn_string, conn_cache[i].b_conn_string))
	return i;
      /* FIXME: free ones not used in a while */
    }

  /* Allocate more space if necessary */
  /* FIXME: check for out of memory */
  if (firstfree == -1)
    {
      if (!conn_cache_size)
	{
	  conn_cache=malloc(sizeof(struct brl_oci_connection));
	  conn_cache_size=1;
	}
      else
	{
	  conn_cache=
	    realloc(conn_cache,
		    ++conn_cache_size * sizeof(struct brl_oci_connection));
	}
      firstfree=conn_cache_size-1;
    }
  conn_cache[firstfree].b_conn_string=strdup(conn_string);
  conn_cache[firstfree].b_lda = malloc(sizeof(Lda_Def));
  conn_cache[firstfree].b_hda = malloc(512);

  if (!strchr(conn_string, '/'))
    {
      int chlen;

      /* We are supposed to autoconnect this user. */
      sprintf(new_password, "%c%x%x", rand()%26 + 'a', rand(), rand());
      /* Note: the extra space after "alter" is deliberate. */
      sprintf(sqlbuf, "ALTER  USER %s IDENTIFIED BY %s",
	      conn_string, new_password);

      /* If we can't alter the user, need to create it. */
      if (brl_oci_dosql("-", sqlbuf, stderr))
	{
	  strncpy(sqlbuf, "CREATE", 6);
	  brl_oci_dosql("-", sqlbuf, stderr);
	  sprintf(sqlbuf, "GRANT CONNECT, BRL_AUTH TO %s", conn_string);
	  brl_oci_dosql("-", sqlbuf, stderr);
	}
      sprintf(sqlbuf, "%s/%s", conn_string, new_password);
      conn_string = sqlbuf;
    }

  /* Connect */
  olog(conn_cache[firstfree].b_lda, conn_cache[firstfree].b_hda,
       (text *)conn_string, (sword)-1, (text *)0, (sword)0,
       (text *)0, (sword)0, OCI_LM_DEF);

  time(&(conn_cache[firstfree].b_lastused));
  return firstfree;
}

int
brl_oci_init(brlbuddy_connstring)
     char *brlbuddy_connstring;
{
  int retval;

  /* This should be the first call to oci_connect, so 0 should be returned */
  retval = brl_oci_connect(brlbuddy_connstring);
  if (retval !=0) return retval;

  /* Wipe master connection string from memory in case we dump core */
  memset(conn_cache[0].b_conn_string, 0,
	 strlen(conn_cache[0].b_conn_string));
  conn_cache[0].b_conn_string[0]='-';
  return conn_cache[0].b_lda->rc;
}

void
brl_oci_writestring(scheme_file, str, len)
     FILE *scheme_file;
     char *str;
     int len;
{
  char *ptr1, *ptr2, ch3[3];

  tstart();

  strcpy(ch3, "\\ ");
  if (len == -1) len = strlen(str);
  fwrite("\"", 1, 1, scheme_file);
  for(ptr1=str; ptr1 < str + len; )
    {
      /* Find quote or backslash */
      ptr2=strpbrk(ptr1, "\"\\");
      if (!ptr2 || ptr2 >= str + len)
	{
	  fwrite(ptr1, 1, len - (ptr1 - str), scheme_file);
	  break;
	}
      ch3[1]=*ptr2;
      fwrite(ptr1, 1, ptr2 - ptr1, scheme_file);
      fwrite(ch3, 1, 2, scheme_file);
      ptr1=ptr2+1;
    }
  fwrite("\"", 1, 1, scheme_file);

  usec_brl_oci_writestring += tend();
  return;
}

void
brl_oci_error(lda, cda, context_string, scheme_file)
     Lda_Def *lda;
     Cda_Def *cda;
     char *context_string;
     FILE *scheme_file;
{
  text message_buffer[1024];

  oerhms(lda, cda->rc, message_buffer, sizeof(message_buffer));
  strcat(strcat(strcat(message_buffer, " ("), context_string), ")\n");
  fwrite("(error ", 1, 7, scheme_file);
  brl_oci_writestring(scheme_file, message_buffer, -1);
  fwrite(")", 1, 1, scheme_file);
  return;
}

int
brl_oci_dosql(conn_string, sql_expr, scheme_file)
     char *conn_string, *sql_expr;
     FILE *scheme_file;
{
  int c, pos, i, retval=0;
  Lda_Def *lda;
  Cda_Def ocursor;
  struct brl_oci_outvar **ov;
  int ncols;
  char rows_processed[32];

#ifndef NO_BRL_DEBUG
  if (debugfd != -1)
    {
      write(debugfd, sql_expr, strlen(sql_expr));
      write(debugfd, "\n", 1);
    }
#endif

  /* Connect */
  c = brl_oci_connect(conn_string);
  if (c == -1)
    {
      fwrite(brl_nomem, 1, strlen(brl_nomem), scheme_file);
      return 1;
    }
  lda = conn_cache[c].b_lda;
  if (lda->rc)
    {
      retval=(int)lda->rc;
      brl_oci_error(lda, lda, conn_string, scheme_file);
      return retval;
    }

  /* Define a cursor */
  tstart();
  retval=oopen(&ocursor, lda, NULL, -1, -1, NULL, -1);
  usec_oopen += tend();
  if (retval)
    {
      retval=(int)lda->rc;
      brl_oci_error(lda, lda, "oopen", scheme_file);
      return retval;
    }

  /* Parse the sql statement */
  tstart();
  retval=oparse(&ocursor, (text *)sql_expr, (sb4) -1, (sword) 1, (ub4) 2);
  usec_oparse += tend();
  if (retval)
    {
      retval=(int)ocursor.rc;
      brl_oci_error(lda, &ocursor, sql_expr, scheme_file);
      oclose(&ocursor);
      return retval;
    }

  /* Allocate pointer to first output variable */
  ov = malloc(sizeof(struct brl_oci_outvar *));

  /* Define output variables for each column */
  for(pos=1; ; pos++)
    {
      /* Allocate this output variable */
      if (pos > 1)
	ov = realloc(ov, pos*sizeof(struct brl_oci_outvar *));
      ov[pos-1] = calloc(1, sizeof(struct brl_oci_outvar));

      /* Get a description of this column */
      ov[pos-1]->b_cbufl = MAX_COLUMN_DESCRIPTION;
      tstart();
      retval=odescr(&ocursor, pos, &(ov[pos-1]->b_dbsize), &(ov[pos-1]->b_dbtype),
		    ov[pos-1]->b_cbuf, &(ov[pos-1]->b_cbufl),
		    &(ov[pos-1]->b_dsize), &(ov[pos-1]->b_prec),
		    &(ov[pos-1]->b_scale), &(ov[pos-1]->b_nullok));
      usec_odescr += tend();
      if (retval)
	{
	  if (ocursor.rc == 1007)
	    {
	      free(ov[pos-1]);
	      break; /* end of description */
	    }
	  else
	    {
	      free(ov[pos-1]);
	      retval=(int)ocursor.rc;
	      brl_oci_error(lda, &ocursor, sql_expr, scheme_file);
	      goto cleanup;
	    }
	}

      /* Define an output variable for this column */
      ov[pos-1]->b_bufl = 2+ov[pos-1]->b_dsize;
      if (ov[pos-1]->b_bufl == 2 || ov[pos-1]->b_bufl > 65535)
	ov[pos-1]->b_bufl = 65535; /* max VARCHAR */
      ov[pos-1]->b_buf = (ub1 *) malloc(ov[pos-1]->b_bufl);
      /* FIXME: check retval */
      ov[pos-1]->b_ftype = 9;	/* VARCHAR */
      tstart();
      odefin(&ocursor, pos, ov[pos-1]->b_buf, ov[pos-1]->b_bufl,
	     ov[pos-1]->b_ftype, ov[pos-1]->b_scale, &(ov[pos-1]->b_ind),
	     (text *)0, (sword)0, (sword)0, &(ov[pos-1]->b_rlen),
	     &(ov[pos-1]->b_rcode));
      usec_odefin += tend();
    }

  /* Execute the statement */
  tstart();
  retval=oexec(&ocursor);
  usec_oexec += tend();
  if (retval)
    {
      retval=(int)ocursor.rc;
      brl_oci_error(lda, &ocursor, sql_expr, scheme_file);
      goto cleanup;
    }

  /* Print the results */
  if (pos == 1)
    {
      /* Non-table results */
      if (fileno(scheme_file) != 2)	/* don't print to stderr */
	{
	  sprintf(rows_processed, "%d\n", ocursor.rpc);
	  fwrite(rows_processed, 1, strlen(rows_processed), scheme_file);
	}
    }
  else
    {
      fwrite("(\n", 1, 2, scheme_file);

      /* Write column headings */
      fwrite("(", 1, 1, scheme_file);
      for(i=1; i < pos; i++)
	{
	  if (i > 1) fwrite(" ", 1, 1, scheme_file);
	  brl_oci_writestring(scheme_file, ov[i-1]->b_cbuf, ov[i-1]->b_cbufl);
	}
      fwrite(")\n", 1, 2, scheme_file);

      while(ocursor.rc == 0)
	{
	  tstart();
	  retval=ofetch(&ocursor);
	  usec_ofetch += tend();
	  if (!retval)
	    {
	      /* Print this row */
	      fwrite("(", 1, 1, scheme_file);
	      for(i=1; i < pos; i++)
		{
		  if (i > 1) fwrite(" ", 1, 1, scheme_file);
		  if (ov[i-1]->b_ind == -1)
		    {
		      /* SQL NULL */
		      fwrite("()", 1, 2, scheme_file);
		    }
		  else
		    {
		      switch(ov[i-1]->b_dbtype)
			{
			case 2:	/* NUMBER */
			case 11:	/* ROWID */
			  fwrite(2+ov[i-1]->b_buf, 1,
				 ov[i-1]->b_rlen, scheme_file);
			  break;
			default:
			  brl_oci_writestring(scheme_file,
					      2+ov[i-1]->b_buf,
					      ov[i-1]->b_rlen);
			  break;
			}
		    }
		}
	      fwrite(")\n", 1, 2, scheme_file);
	    }
	}
      fwrite(")\n", 1, 2, scheme_file);
    }

cleanup:
  while(pos > 0)
    {
      if (ov[pos-1]->b_buf) free(ov[pos-1]->b_buf);
      free(ov[--pos]);
    }
  free(ov);
  oclose(&ocursor);
  time(&(conn_cache[c].b_lastused));
  return retval;
}

#endif /* HAVE_LIBCLIENT - OCI client library */
