(define (brl-read-all)
  (read-string (char-set)))

(define (brl-unix-output cmd input-str)
  (let ((infilename (->namestring (temporary-file-pathname)))
	(outfilename (->namestring (temporary-file-pathname)))
	(thunk (lambda () (display input-str))))
    (begin
      (with-output-to-file infilename thunk)
      (unix/system (string-append cmd " < " infilename " > " outfilename))
      (let ((result (with-input-from-file outfilename brl-read-all)))
	(delete-file-no-errors infilename)
	(delete-file-no-errors outfilename)
	result))))

;;;
;;; MIME support: Use quoted-printable encoding for 8-bit chars
;;;
(define (brl-qp-encode str)
  (brl-unix-output "mmencode -q" str))

;;;
;;; PGP support: Encrypt with ASCII armor for e-mail
;;; (brl-pgp-encrypt userid str) ==> str PGP-encrypted for userid
;;; (brl-pgp-encrypt userid) ==> procedure to PGP-encrypt str for userid
;;;

(define brl-pgp-encrypt
  (lambda args
    (cond ((= 2 (length args))
	   ; Called with args for
	   (brl-unix-output (string-append "pgp -feat "
					   (car args))
			    (cadr args)))
	  ((= 1 (length args))
	   (lambda (str) (brl-pgp-encrypt (car args) str)))
	  (else (error "Procedure pgp requires 1 or 2 args")))))
	   

(define (brl-mail-string str)
  (brl-unix-output "/usr/lib/sendmail -oi -t" (brl-unix-eol str))
  str)

(define (brl-mail-string-debug str)
  (brl-unix-output "/usr/lib/sendmail -oi -odi -t" (brl-unix-eol str))
  str)

;;;
;;; Unix newline convention: LF
;;; Convert CR/LF or CR to LF.
;;;

(define brl-cr (integer->char 13))
(define brl-lf (integer->char 10))

(define (brl-unix-eol-convert chl-from chl-to-reversed)
  (cond ((null? chl-from) chl-to-reversed)
	((equal? (car chl-from) brl-cr)
	 (cond ((null? (cdr chl-from))
		(cons brl-lf chl-to-reversed))
	       ((equal? brl-lf (cadr chl-from))
		(brl-unix-eol-convert (cdr chl-from) chl-to-reversed))
	       (else
		(brl-unix-eol-convert
		 (cdr chl-from) (cons brl-lf chl-to-reversed)))))
	(else
	 (brl-unix-eol-convert (cdr chl-from)
			       (cons (car chl-from) chl-to-reversed)))))

(define (brl-unix-eol str)
  (list->string (reverse (brl-unix-eol-convert (string->list str) '()))))
